@@ -16,4 +16,17 @@ class Error(Exception):
         self.rule = rule
 
     def __str__(self):
-        return self.error
+        return self.error or ''
+
+
+class ParseError(Error):
+    """Could not parse a rule."""
+
+class EvalError(Error):
+    """Could not evaluate a rule."""
+
+class ValidationError(Error):
+    """Could not validate a field."""
+
+class MissingFieldError(Error):
+    """A mandatory field is missing."""
